const Clutter = imports.gi.Clutter;
const GObject = imports.gi.GObject;
const Gdk     = imports.gi.Gdk;
const Gtk     = imports.gi.Gtk;
const St      = imports.gi.St;

const DROP_AFTER = Gtk.TreeViewDropPosition.AFTER;

const LEFT  = true;
const RIGHT = false;

function animation(left) {
    let monitorWidth = imports.ui.main.layoutManager.primaryMonitor.width;

    let corner = new St.Bin({
        style_class : 'dot',
        track_hover : true,
        can_focus   : true,
        reactive    : true,
        x_fill      : true,
        y_fill      : true,
        opacity     : 0
    });

    let ripples = [];
    for (let i = 0; i < 3; ++i) {
        ripples[i] = new St.BoxLayout({
            style_class : left ? 'norm-ripple' : 'inv-ripple',
            visible     : false,
            opacity     : 0
        });
    }

    function animate(corner, gravity, ripple,
                     delay, time, startScale, startOpacity, finalScale) {
        ripple._opacity = startOpacity;
        ripple.set_anchor_point_from_gravity(gravity);

        ripple.visible = true;
        ripple.opacity = 255 * Math.sqrt(startOpacity);
        ripple.scale_x = ripple.scale_y = startScale;

        let [ x, y ] = corner.get_transformed_position();
        ripple.x = x;
        ripple.y = y;

        imports.ui.tweener.addTween(ripple, {
            _opacity: 0,
            scale_x: finalScale,
            scale_y: finalScale,
            delay: delay,
            time: time,
            transition: 'linear',
            onUpdate : function() {
                ripple.opacity = 255 * Math.sqrt(ripple._opacity);
            },
            onComplete : function() {
                ripple.visible = false;
            }
        });
    };

    return [ corner, ripples, {
        show : function() {
            let delays         = [ 0.00, 0.05, 0.35 ];
            let times          = [ 0.83, 1.00, 1.00 ];
            let startScales    = [ 0.25, 0.00, 0.00 ];
            let startOpacities = [ 1.00, 0.70, 0.30 ];
            let finalScales    = [ 1.50, 1.25, 1.00 ];

            let gravity;
            let ltr = corner.get_text_direction() == Clutter.TextDirection.LTR;

            if (left == ltr) {
                corner.set_position(0, 0);
                gravity = Clutter.Gravity.NORTH_WEST;
            } else {
                corner.set_position(monitorWidth - 1, 0);
                gravity = Clutter.Gravity.NORTH_EAST;
            }

            for (let i = 0; i < ripples.length; ++i) {
                animate(corner, gravity, ripples[i],
                        delays[i], times[i], startScales[i], startOpacities[i], finalScales[i]);
            }
        }
    } ];
}

function textColumn() {
    let text = new Gtk.CellRendererText();

    let column = new Gtk.TreeViewColumn();
    column.pack_start(text, true);
    column.add_attribute(text, 'text', 1);

    return column;
}

function accelColumn() {
    let accel = new Gtk.CellRendererAccel({
        editable   : true,
        accel_mode : Gtk.CellRendererAccelMode.GTK
    });

    let column = new Gtk.TreeViewColumn();
    column.pack_start(accel, false);
    column.add_attribute(accel, 'accel-mods', 2);
    column.add_attribute(accel, 'accel-key', 3);

    return [ accel, column ];
}

function treeView(model, reorderable) {
    return new Gtk.TreeView({
        model           : model,
        headers_visible : false,
        enable_search   : true,
        reorderable     : reorderable
    });
}

function accelView(init) {
    let model = new Gtk.ListStore();
    model.set_column_types([ GObject.TYPE_STRING, GObject.TYPE_STRING,
                             GObject.TYPE_INT, GObject.TYPE_INT ]);

    let [ accel, column ] = accelColumn();
    init(model, accel);

    let tree = treeView(model, false);
    tree.append_column(textColumn());
    tree.append_column(column);

    return tree;
}

function dndView(init, forbidDrop, forbidDrag) {
    let model = new Gtk.TreeStore();
    model.set_column_types([ GObject.TYPE_STRING, GObject.TYPE_STRING ]);
    init(model);

    let tree = treeView(model, true);
    tree.append_column(textColumn());
    tree.drag_dest_set_track_motion(true);

    let scroll = new Gtk.ScrolledWindow();
    scroll.add(tree);

    function consume(x, y) {
        let [ source, column ] = tree.get_cursor();
        let [ ok, dest, drop ] = tree.get_dest_row_at_pos(x, y);

        return !ok || source == null || forbidDrop(model, source, dest, drop);
    }

    tree.connect('drag-drop', function(tree, context, x, y, time) {
        return consume(x, y);
    });

    tree.connect('drag-motion', function(tree, context, x, y, time) {
        return consume(x, y);
    });

    tree.connect('drag-data-get', function(tree, context, data, info, time) {
        let [ path, column ] = tree.get_cursor();
        if (path && forbidDrag(model, path)) {
            Gdk.drag_abort(context, time);
        }
    });

    return scroll;
}

function sidebarView(init, select) {
    let model = new Gtk.ListStore();
    model.set_column_types([ GObject.TYPE_STRING, GObject.TYPE_STRING ]);
    init(model);

    let tree = treeView(model, false);
    tree.append_column(textColumn());

    let box = new Gtk.VBox();

    let pan = new Gtk.HPaned({
        expand : true
    });

    pan.pack1(tree, false, false);
    pan.pack2(box, true, true);

    tree.connect('cursor-changed', function(tree) {
        let [ path, column ] = tree.get_cursor();
        if (path) {
            let [ ok, iter ] = model.get_iter(path);
            if (ok) {
                box.get_children().forEach(function(widget, i, a) {
                    box.remove(widget);
                });
                box.add(select(model.get_value(iter, 0)));
                box.show_all();
            }
        }
    });

    return pan;
}

function switchOption(label, active, toggle) {
    let left = new Gtk.HBox({
        halign : Gtk.Align.START
    });
    left.add(new Gtk.Label({
        label : label,
        wrap  : true
    }));

    let right = new Gtk.HBox({
        halign : Gtk.Align.END,
        valign : Gtk.Align.CENTER
    });

    let option = new Gtk.Switch({
        active : active()
    });
    right.add(option);

    let box = new Gtk.HBox({
        hexpand : true
    });

    box.add(left);
    box.add(right);

    option.connect('notify::active', toggle);

    return box;
}

function grid() {
    let grid = new Gtk.Grid({
        orientation : Gtk.Orientation.VERTICAL
    });

    for each (let widget in arguments) {
        grid.add(widget);
    }

    return grid;
}
